#!/usr/bin/env python

import sys

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from simple_lexer   import Lexer
from simple_grammar import Grammar, Rule, Expression, Alternative, Ebnf, Nonterminal, Terminal
from simple_symbols import initSymbols

# --------------------------------------------------------------------------

class Highlighter (QSyntaxHighlighter) :
    def __init__ (self, parent = None) :
        super (Highlighter, self).__init__ (parent)
        self.lexer = Lexer ()

        self.identifierFormat = QTextCharFormat ()
        self.identifierFormat.setForeground (Qt.darkRed)
        self.identifierFormat.setToolTip ("identifier")

        self.numberFormat = QTextCharFormat ()
        self.numberFormat.setForeground (Qt.cyan)

        self.realFormat = QTextCharFormat ()
        self.realFormat.setForeground (Qt.magenta)

        self.characterFormat = QTextCharFormat ()
        self.characterFormat.setForeground (QColor ("cornflowerblue"))

        self.stringFormat = QTextCharFormat ()
        self.stringFormat.setForeground (Qt.blue)

        self.separatorFormat = QTextCharFormat ()
        self.separatorFormat.setForeground (QColor ("orange"))

        self.commentFormat = QTextCharFormat ()
        self.commentFormat.setForeground (Qt.gray)

    def highlightBlock (self, text) :
        #try :
           lexer = self.lexer
           lexer.openString (text)
           while not lexer.isEndOfSource () :
              index = lexer.tokenByteOfs
              length = lexer.byteOfs - lexer.tokenByteOfs
              if lexer.token == lexer.identifier :
                 self.setFormat (index, length, self.identifierFormat)
              elif lexer.token == lexer.number :
                 self.setFormat (index, length, self.numberFormat)
              elif lexer.token == lexer.real :
                 self.setFormat (index, length, self.realFormat)
              elif lexer.token == lexer.character :
                 self.setFormat (index, length, self.characterFormat)
              elif lexer.token == lexer.string :
                 self.setFormat (index, length, self.stringFormat)
              elif lexer.token == lexer.separator :
                 self.setFormat (index, length, self.separatorFormat)
              else : # !?
                 self.setFormat (index, length, self.commentFormat)
              lexer.nextToken ()
        #except :
              #pass

# --------------------------------------------------------------------------

class Edit (QTextEdit) :
    def __init__ (self, parent=None) :
        super (Edit, self).__init__ (parent)
        self.highlighter = Highlighter (self.document ())

    def selectLine (self, line) :
        cursor = self.textCursor ()
        cursor.movePosition (QTextCursor.Start)
        cursor.movePosition (QTextCursor.Down, QTextCursor.MoveAnchor, line-1)
        cursor.movePosition (QTextCursor.EndOfLine, QTextCursor.KeepAnchor)
        self.setTextCursor (cursor)

    def select (self, line, col, stop_line, stop_col):
        cursor = self.textCursor ()

        cursor.movePosition (QTextCursor.Start)
        cursor.movePosition (QTextCursor.Down, QTextCursor.MoveAnchor, line-1)
        cursor.movePosition (QTextCursor.Right, QTextCursor.MoveAnchor, col-1)

        if stop_line == line :
           cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, stop_col-col)
        else :
           cursor.movePosition (QTextCursor.Down, QTextCursor.KeepAnchor, stop_line-line)
           cursor.movePosition (QTextCursor.StartOfLine, QTextCursor.KeepAnchor)
           cursor.movePosition (QTextCursor.Right, QTextCursor.KeepAnchor, stop_col)

        self.setTextCursor (cursor)

# --------------------------------------------------------------------------

def findIcon (icon_name) :
    return QIcon.fromTheme (icon_name)

class TreeItem (QTreeWidgetItem) :
    def __init__ (self, parent, text) :
        super (TreeItem, self).__init__ (parent)
        self.setText (0, text)

    def addIcon (self, icon_name) :
        self.setIcon (0, findIcon (icon_name))

class TreeView (QTreeWidget) :
    def __init__ (self, parent=None) :
        super (TreeView, self).__init__ (parent)
        self.setColumnCount (1)
        self.header ().hide ()
        self.setEditTriggers (QAbstractItemView.SelectedClicked)
        self.setIndentation (8)

# --------------------------------------------------------------------------

class MainWin (QMainWindow) :

    def __init__ (self, parent=None) :
        super (MainWin, self).__init__ (parent)
        self.setupUI ()
        self.setupMenus ()

    def setupUI (self) :
        self.resize (640, 480)

        self.split = QSplitter ()
        self.setCentralWidget (self.split)

        self.tree = TreeView ()
        self.split.addWidget (self.tree)

        self.tree.itemActivated.connect (self.onItemActivated)

        self.edit = Edit ()
        self.split.addWidget (self.edit)
        self.edit.setLineWrapMode (QTextEdit.NoWrap) # otherwise selectLine is not exact

    def setupMenus (self) :

        fileMenu = self.menuBar().addMenu ("&File")

        act = QAction ("&Open...", self)
        act.triggered.connect (self.openFile)
        fileMenu.addAction (act)

        act = QAction ("&Quit", self)
        act.triggered.connect (self.close)
        fileMenu.addAction (act)

    def onItemActivated (self, item, column) :
        if hasattr (item, "line") :
           self.edit.selectLine (item.line)

    def openFile (self) :
        fileName = QFileDialog.getOpenFileName (self, "Open File")
        if fileName :
           self.compileFile (str (fileName))

    def compileFile (self, fileName) :
        self.tree.clear ()
        self.edit.clear ()

        grammar = Grammar ()
        grammar.openFile (fileName)
        self.edit.setPlainText (grammar.source)

        grammar.parseRules ()
        initSymbols (grammar)

        self.grammar = grammar
        self.addSymbols ()
        for rule in grammar.rules :
           self.addBranch (self.tree, rule)

    def addSymbols (self) :
        branch = TreeItem (self.tree, "symbols")
        for symbol in self.grammar.symbols :
            TreeItem (branch, str (symbol.inx) + " " + symbol.name)

    def addBranch (self, above, node) :

        txt = ""
        if isinstance (node, Rule) :
           txt = node.name
        elif isinstance (node, Expression) :
           txt = "expression"
        elif isinstance (node, Alternative) :
           txt = "alternative"
        elif isinstance (node, Ebnf) :
           txt = "ebnf " + node.mark
        elif isinstance (node, Nonterminal) :
           txt = "nonterminal "
           if node.variable != "" :
              txt = txt + node.variable + ":"
           txt = txt + node.rule
        elif isinstance (node, Terminal) :
           txt = "terminal " + node.text
        else:
           txt = node.__class__.__name__

        item = TreeItem (above, txt)

        if isinstance (node, Rule) :
           if hasattr (node, "line") :
              item.line = node.line
           item.addIcon ("code-class")
        if isinstance (node, Nonterminal) :
           item.addIcon ("code-function")
        if isinstance (node, Terminal) :
           item.addIcon ("code-variable")
        if isinstance (node, Ebnf) :
           item.addIcon ("code-block")

        if isinstance (node, Rule) :
           if hasattr (node, "first") :
              n = 0
              for inx in range (len (node.first)) :
                  if node.first [inx] :
                     n = n + 1
              if n > 1 :
                 item.setForeground (0, QBrush (Qt.green))

        if hasattr (node, "nullable") :
           if node.nullable :
              item.setForeground (0, QBrush (Qt.red))

        if isinstance (node, Rule) :
           self.addInfo (item, node)
           self.addBranch (item, node.expr)
        elif isinstance (node, Expression) :
           for t in node.alternatives :
              self.addBranch (item, t)
        elif isinstance (node, Alternative) :
           for t in node.items :
              self.addBranch (item, t)
        elif isinstance (node, Ebnf) :
           self.addBranch (item, node.expr)
           if node.impl != None :
              self.addBranch (item, node.impl)

    def addInfo (self, above, rule) :
        if hasattr (rule, "first") :
           branch = TreeItem (above, "first")
           branch.addIcon ("info")
           branch.setForeground (0, QBrush (Qt.blue))
           for inx in range (len (rule.first)) :
               if rule.first [inx] :
                  name = self.grammar.symbols [inx].name
                  item = TreeItem (branch, name)
                  item.setForeground (0, QBrush (Qt.blue))

# --------------------------------------------------------------------------

app = QApplication (sys.argv)
QIcon.setThemeName ("oxygen")
win = MainWin ()
# win.edit.setFont (QFont ("Sans Serif", 24))
# win.tree.setFont (win.edit.font ())
win.show ()
win.compileFile ("pas.g")
app.exec_ ()

# kate: indent-width 1; show-tabs true; replace-tabs true; remove-trailing-spaces all
